// Carl Ebeling and Steve Gribble (based on Stuart Reges' code)
// 3/01/2006
//
// Class CritterModel keeps track of the state of the critter simulation.

import java.util.*;
import java.awt.*;

public class CritterModel {
    private static int myHeight;
    private static int myWidth;
    private static Critter[][] myGrid, newGrid;
    private Map<Critter, State> myList;
    private boolean mySimpleWorld;
    private int iterationsPerGame;

    private static final int MAX_ITERATIONS = 100;
    private static final int MIN_ITERATIONS = 5;

    public CritterModel(boolean simpleWorld, int width, int height) {
        myWidth = width;
        myHeight = height;
        newGrid = myGrid = new Critter[width][height];
        myList = new HashMap<Critter, State>();
        mySimpleWorld = simpleWorld;

        Random r = new Random();
        iterationsPerGame = MIN_ITERATIONS +
            r.nextInt(1+MAX_ITERATIONS-MIN_ITERATIONS);
        
    }

    public void add(int number, Class critter) {
        if (myList.size() + number > myWidth * myHeight)
            throw new RuntimeException("adding too many critters");
        for (int i = 0; i < number; i++) {
            Critter next;
            try {
                next = (Critter) critter.newInstance();
            } catch (Exception e) {
                throw new RuntimeException("" + e);
            }
            int x, y;
            do {
                x = randomInt(0, myWidth - 1);
                y = randomInt(0, myHeight - 1);
            } while (myGrid[x][y] != null);
            myGrid[x][y] = next;
            myList.put(next, new State(x, y, 0));
        }
    }

    private static int randomInt(int low, int high) {
        return low + (int)(Math.random() * (high - low + 1));
    }

    public int getWidth() {
        return myWidth;
    }

    public int getHeight() {
        return myHeight;
    }

    public char getChar(int x, int y) {
        if (myGrid[x][y] == null)
            return '.';
        else {
            char c = myGrid[x][y].getChar();
            return c;
        }
    }
    
    // Turn money into color
    //   BLUE:   richest critter(s)
    //   RED:    poorest critter(s)
    //   BLACK:  everyone else
    //
    public Color getColor(int x, int y) {

        if (myGrid[x][y] == null)
            return Color.BLACK;

        int richest = -1, poorest = java.lang.Integer.MAX_VALUE;
        Object[] list = myList.keySet().toArray();

        for (int i=0; i<list.length; i++) {
            Critter next = (Critter) list[i];
            State p = myList.get(next);

            if (p.money > richest)
                richest = p.money;
            if (p.money < poorest)
                poorest = p.money;
        }
        State p = myList.get(myGrid[x][y]);
        if (p.money == richest)
            return Color.BLUE;
        else if (p.money == poorest)
            return Color.RED;
        else
            return Color.BLACK;
    }

    // Used to give wolves an extra turn
    static boolean wolfTurn = false;
    // Note that multiple animals can be on a location at
    // once.  Only one of the animals appears in the grid
    // array, however, and that is the animal that is
    // visible.
    public void update() {
        newGrid = new Critter[myWidth][myHeight];
        Object[] list = myList.keySet().toArray();
        Collections.shuffle(Arrays.asList(list));

        for (int i = 0; i < list.length; i++) {
            Critter next = (Critter) list[i];
            State p = myList.get(next);
            if (wolfTurn && !(next.getClass().getSimpleName().equals("Wolf"))) {
                if (newGrid[p.x][p.y] == null) {
                    newGrid[p.x][p.y] = next;
                }
                continue;
            }
            int move = next.getMove(p.x, p.y);
            if (move == Critter.NORTH)
                p.y = (p.y + myHeight - 1) % myHeight;
            else if (move == Critter.SOUTH)
                p.y = (p.y + 1) % myHeight;
            else if (move == Critter.EAST)
                p.x = (p.x + 1) % myWidth;
            else if (move == Critter.WEST)
                p.x = (p.x + myWidth - 1) % myWidth;
            else
                throw new RuntimeException("Illegal move from " + list[i]);

            if (mySimpleWorld) {
                if (newGrid[p.x][p.y] != null) {
                    myList.remove(newGrid[p.x][p.y]);
                }
                newGrid[p.x][p.y] = next;
            } else {
                // complexworld rules
                if (newGrid[p.x][p.y] != null) {

                    // play the game with whomever ended up on grid last
                    playGame(next, newGrid[p.x][p.y]);
                }
                newGrid[p.x][p.y] = next;
            }

        }
        myGrid = newGrid;
        wolfTurn = !wolfTurn;	// Alternate wolf-only turns
    }

    public void dumpMoney() {
        Object[] list = myList.keySet().toArray();

	System.out.println("");
        for (int i=0; i<list.length; i++) {
            Critter next = (Critter) list[i];
            State p = myList.get(next);

            System.out.println(next.getClass().getSimpleName() + 
                               " at (" + p.x + "," + p.y + 
                               ") has $" + p.money + ".");
        }
    }

    private void playGame(Critter first,  Critter second) {
        State state1 = myList.get(first);
        State state2 = myList.get(second);

        int last_coop1 = -1, last_coop2 = -1;
        for (int i=0; i<iterationsPerGame; i++) {
            boolean coop1, coop2;

            coop1 = first.decide(last_coop2, second.getClass());
            coop2 = second.decide(last_coop1, first.getClass());

            last_coop1 = (coop1 ? 1 : 0);
            last_coop2 = (coop2 ? 1 : 0);

            if (coop1 && coop2) {
                state1.money += 2;
                state2.money += 2;
            } else if (!coop1 && !coop2) {
                state1.money += 1;
                state2.money += 1;
            } else if (coop1 && !coop2) {
                state1.money += 0;
                state2.money += 3;
            } else {
                state1.money += 3;
                state2.money += 0;
            }
        }
    }

    // This first attempts to use newGrid, so it sees the critters that
    // have already moved.  If newGrid is empty, it then tries myGrid,
    // so it sees critters that haven't yet moved.
    //
    // Bug: if a critter did move, you might see it's "ghost" in myGrid.
    public static Class neighbor(int x, int y, int dx, int dy) {
        if (dx > 1) dx = 1;
        if (dx < -1) dx = -1;
        if (dy > 1) dy = 1;
        if (dy < -1) dy = -1;
        return oneNeighbor(x, y, dx, dy);
    }

    private static Class oneNeighbor(int x, int y, int xInc, int yInc) {
        x = (x + xInc + myWidth) % myWidth;
        y = (y + yInc + myHeight) % myHeight;
        if (newGrid[x][y] == null) {
            if (myGrid[x][y] == null)
                return null;
            else
                return myGrid[x][y].getClass();
        } else
            return newGrid[x][y].getClass();
    }
}
